#include	<windows.h>

#include	"common.h"
#include	"i286.h"
#include	"memory.h"
#include	"iobridge.h"
#include	"dmac.h"
#include	"fdc.h"
#include	"nevent.h"
#include	"pic.h"
#include	"trace.h"

	DMA_TABLE	dma;

static void __fastcall dummy_out(BYTE data) { }
static BYTE __fastcall dummy_in(void) { return(0xff); }
static void dummy_end(void) { }

void dmac_init(void) {

	int		i;

	ZeroMemory(&dma, sizeof(dma));
	dma.mask = 0xf;
	for (i=0; i<4; i++) {
		dma.dmach[i].outproc = dummy_out;
		dma.dmach[i].inproc = dummy_in;
		dma.dmach[i].endproc = dummy_end;
	}
	dma.dmach[DMA_FDD].inproc = FDC_DataRegRead;
	dma.dmach[DMA_FDD].endproc = fdc_tc;
}

void dmac_check(void) {

	int		i;
	int		work = 0;
	BYTE	bit;

	for (i=0, bit=1; i<4; i++, bit<<=1) {
		if ((!(dma.mask & bit)) && (dma.dmach[i].ready)) {
			work |= bit;
		}
	}
	if (work != dma.work) {
		if ((work == 0) != (dma.work == 0)) {
			nevent_forceexit();
		}
		dma.work = work;
	}
}


I286_OUT dmac_w(WORD port, BYTE value) {

	if (!(port & 1)) {
		return;
	}
	TRACEOUT(port, value);
	if (port < 0x10) {
		if (port & 2) {
			dma.dmach[port >> 2].leng.b[dma.lh] = value;
			dma.dmach[port >> 2].lengbak.b[dma.lh] = value;
		}
		else {
			dma.dmach[port >> 2].adrs.b[dma.lh] = value;
			dma.dmach[port >> 2].adrsbak.b[dma.lh] = value;
		}
		dma.lh ^= 1;
	}
	else {
		switch(port) {
			case 0x11:
				break;
			case 0x13:
				dma.dmach[value & 3].sreq = value;
				break;
			case 0x15:
				if (value & 4) {
					dma.mask |= (1 << (value & 3));
				}
				else {
					dma.mask &= ~(1 << (value & 3));
				}
				dmac_check();
				break;
			case 0x17:
				dma.dmach[value & 3].mode = value;
				break;
			case 0x19:
				dma.lh = 0;
				break;
			case 0x1b:
				dma.mask = 0x0f;
				break;
			case 0x1d:
				break;
			case 0x1f:
				dma.mask = value;
				dmac_check();
				break;
		}
	}
}

I286_INB dmac_r(WORD port) {

	BYTE	ret;

	if (port & 0x01) {
		TRACEIN(port);
		if (port < 0x10) {
			if (port & 2) {
				ret = dma.dmach[port >> 2].adrs.b[dma.lh];
			}
			else {
				ret = dma.dmach[port >> 2].leng.b[dma.lh];
			}
			dma.lh ^= 1;
			return(ret);
		}
	}
	return(0xff);
}


I286_OUT dmabank_w(WORD port, BYTE value) {

	if (port & 1) {
		if (port < 0x28) {
			dma.dmach[((port + 2) >> 1) & 3].bank.b[0] = value;
		}
		else if (port == 0x29) {
		}
	}
}

I286_INB dmabank_r(WORD port) {

	return(0xff);
}


LABEL void dma_proc(void) {

		__asm {
				cmp		dma.work, 0
				je		noworkdma
				mov		edi, offset (dma.dmach)
				mov		cl, 4
				mov		ebx, 1
searchworkdma:	test	dma.work, ebx
				jne		hitworkdma
				add		ebx, ebx
				add		edi, type DMACH_TABLE
				dec		cl
				jne		searchworkdma
noworkdma:		ret

hitworkdma:		sub		word ptr ([edi].leng.w), 1
				jnc		dmaprocs
				not		ebx
				and		dma.work, ebx
				call	[edi].endproc

dmaprocs:		mov		al, [edi].mode
				and		al, 0ch
				je		dma_verify
				cmp		al, 04h
				je		dma_port2mem

				mov		ecx, dword ptr ([edi].adrs)
				call	i286_memoryread
				mov		cl, al
				call	[edi].outproc
				jmp		dma_counters

dma_verify:		call	[edi].inproc
				jmp		dma_counters

dma_port2mem:	call	[edi].inproc
				mov		ecx, dword ptr ([edi].adrs)
				mov		dl, al
				call	i286_memorywrite

dma_counters:	test	[edi].mode, 20h
				jne		short dmaadrsdec
				inc		word ptr ([edi].adrs.w)
				ret
dmaadrsdec:		dec		word ptr ([edi].adrs.w)
				ret
	}
}
